import os
import logging
import json
import subprocess
from retrying import retry

from scrapinghub import APIError

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def resolve_project_id(project_id=None):
    """
    Gets project id from following sources in following order of precedence:
    - default parameter values
    - environment variables
    - sh_scrapy.hsref (kumo)
    - scrapinghub.yml file

    in order to allow to use codes that needs HS or dash API,
    either locally or from scrapinghub, correctly configured
    """
    if project_id:
        return project_id

    # read from environment
    if os.environ.get('PROJECT_ID'):
        return os.environ.get('PROJECT_ID')

    # for ScrapyCloud jobs:
    if os.environ.get('SHUB_JOBKEY'):
        return os.environ['SHUB_JOBKEY'].split('/')[0]

    # read from scrapinghub.yml
    try:
        from shub.config import load_shub_config
        cfg = load_shub_config()
        project_id = cfg.get_project_id('default')
        if project_id:
            return project_id
    except:
        logger.warning("Install shub package if want to access scrapinghub.yml")

    if not project_id:
        logger.warning('Project id not found. Use either PROJECT_ID env. variable or scrapinghub.yml default target.')


MINS_IN_A_DAY = 24 * 60
ONE_MIN_IN_S = 60


def just_log_exception(exception):
    logger.error(str(exception))
    for etype in (KeyboardInterrupt, SystemExit, ImportError):
        if isinstance(exception, etype):
            return False
    logger.info("Waiting %d seconds", ONE_MIN_IN_S)
    return True  # retries any other exception


dash_retry_decorator = retry(retry_on_exception=just_log_exception, wait_fixed=ONE_MIN_IN_S*1000,
                             stop_max_attempt_number=MINS_IN_A_DAY)


@dash_retry_decorator
def schedule_script_in_dash(shubproject, cmd, tags=None, units=None, meta=None):
    """
    :type shubproject: scrapinghub.Project
    shubproject - an instance of python-scrapinghub project
    cmd - command line list. First element is script name and rest are options and arguments
    tags - a list of tags to be added to the job
    units - how many Kumo units assign to the job
    meta - a dict with metadata to add to job metadata
    """
    tags = tags or []
    meta = json.dumps(meta) if meta else None
    # lock = str(hash(tuple(cmd))) lock is still not a feature of dash schedule api
    scriptname = cmd[0]
    if not scriptname.startswith('py:'):
        scriptname = 'py:' + scriptname
    try:
        cmd_args = subprocess.list2cmdline(cmd[1:])
        schedule_kwargs = dict(
            spider=scriptname, cmd_args=cmd_args, add_tag=tags, units=units, meta=None)
        logger.info("Scheduling script:\n%s", schedule_kwargs)
        return shubproject.jobs.run(**schedule_kwargs)
    except APIError as e:
        raise RuntimeError(
            "Error scheduling script %s in %s: %s",
            scriptname, shubproject.id, str(e),
        )


def kumo_settings():
    if os.environ.get('SHUB_SETTINGS'):
        from sh_scrapy.env import decode_uri
        return decode_uri(os.environ.get('SHUB_SETTINGS')).get('project_settings', {})
    logging.info("Couldn't find Dash project settings, probably not running in Dash")
    return {}



"""
Implements common methods for ScrapyCloud scripts.
"""
import os
import abc
import logging

from argparse import ArgumentParser

from scrapinghub import ScrapinghubClient

from .utils import (
    resolve_project_id,
    dash_retry_decorator,
)


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.basicConfig(format='%(asctime)s [%(levelname)s] %(message)s')


class BaseScript(abc.ABC):

    name = None  # optional, may be needed for some applications
    flow_id_required = True  # if True, script can only run in the context of a flow_id

    def __init__(self):
        self.project_id = None
        self.client = ScrapinghubClient()
        self.args = self.parse_args()

    def set_flow_id(self, args, default=None):
        self._flow_id = args.flow_id or self.get_flowid_from_tags() or default
        if self.flow_id_required:
            assert not self.flow_id_required or self.flow_id, "Could not detect flow_id. Please set with --flow-id."
        if self.flow_id:
            self.add_job_tags(tags=[f'FLOW_ID={self.flow_id}'])

    @property
    def description(self):
        return "You didn't set description for this script. Please set description property accordingly."

    @property
    def flow_id(self):
        return self._flow_id

    def add_argparser_options(self):
        self.argparser.add_argument('--project-id', help='Overrides target project id.', type=int)
        self.argparser.add_argument('--name', help='Script name.')
        self.argparser.add_argument('--flow-id', help='If given, use the given flow id.')

    def parse_project_id(self, args):
        return args.project_id

    def parse_args(self):
        self.argparser = ArgumentParser(self.description)
        self.add_argparser_options()
        args = self.argparser.parse_args()

        self.project_id = resolve_project_id(self.parse_project_id(args))
        if not self.project_id:
            self.argparser.error('Project id not provided.')

        self.set_flow_id(args)

        self.name = args.name or self.name
        return args

    def get_project(self, project_id=None):
        return self.client.get_project(project_id or self.project_id)

    @dash_retry_decorator
    def get_job_metadata(self, jobid=None, project_id=None):
        """If jobid is None, get own metadata
        """
        jobid = jobid or os.getenv('SHUB_JOBKEY')
        if jobid:
            project = self.get_project(project_id)
            job = project.jobs.get(jobid)
            return job.metadata
        else:
            logger.warning('SHUB_JOBKEY not set: not running on ScrapyCloud.')

    def get_job_tags(self, jobid=None, project_id=None):
        """If jobid is None, get own tags
        """
        metadata = self.get_job_metadata(jobid, project_id)
        if metadata:
            return dict(metadata.list()).get('tags', [])
        return []

    def add_job_tags(self, jobid=None, project_id=None, tags=None):
        """If jobid is None, add tags to own list of tags.
        """
        if tags:
            update = False
            job_tags = self.get_job_tags(jobid, project_id)
            for tag in tags:
                if tag not in job_tags:
                    if tag.startswith('FLOW_ID='):
                        job_tags.insert(0, tag)
                    else:
                        job_tags.append(tag)
                    update = True
            if update:
                metadata = self.get_job_metadata(jobid, project_id)
                if metadata:
                    metadata.update({'tags': job_tags})

    def get_flowid_from_tags(self, jobid=None, project_id=None):
        """If jobid is None, get flowid from own tags
        """
        for tag in self.get_job_tags(jobid, project_id):
            if tag.startswith('FLOW_ID='):
                return tag.replace('FLOW_ID=', '')

    @abc.abstractmethod
    def run(self):
        pass



"""
Implements common methods for any workflow manager
"""
import time
import logging
from uuid import uuid4

from scrapinghub import DuplicateJobError

from .utils import (
    schedule_script_in_dash,
    dash_retry_decorator,
)

from .script import BaseScript


logger = logging.getLogger(__name__)


class WorkFlowManager(BaseScript):

    default_max_jobs = float('inf')
    loop_mode = 0

    def __init__(self):
        self.workflow_loop_enabled = False
        super().__init__()

    def set_flow_id(self, args, default=None):
        default = default or str(uuid4())
        super().set_flow_id(args, default)

    @property
    def max_running_jobs(self):
        return self.args.max_running_jobs

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument('--loop-mode', help='If provided, manager will run in loop mode, with a cycle\
                                    each given number of seconds. Default: %(default)s', type=int, metavar='SECONDS',
                                    default=self.loop_mode)
        self.argparser.add_argument('--max-running-jobs', type=int, default=self.default_max_jobs,
                                    help='If given, don\'t allow more than the given jobs running at once.\
                                    Default: %(default)s')
        self.argparser.add_argument('--tag', help='Add given tag to the scheduled jobs. Can be given multiple times.',
                                    action='append', default=[])

    def _make_tags(self, tags):
        tags = tags or []
        tags.extend(self.args.tag)
        tags.append(f'FLOW_ID={self.flow_id}')
        return list(set(tags)) or None

    def schedule_script(self, cmd, tags=None, project_id=None, **kwargs):
        """
        Schedules an external script
        """
        logger.info('Starting: {}'.format(cmd))
        project = self.get_project(project_id)
        job = schedule_script_in_dash(project, [str(x) for x in cmd], tags=self._make_tags(tags), **kwargs)
        logger.info(f"Scheduled script job {job.key}")
        return job.key

    @dash_retry_decorator
    def schedule_spider(self, spider, tags=None, units=None, project_id=None, **spiderargs):
        schedule_kwargs = dict(spider=spider, add_tag=self._make_tags(tags), units=units, **spiderargs)
        logger.info("Scheduling a spider:\n%s", schedule_kwargs)
        try:
            project = self.get_project(project_id)
            job = project.jobs.run(**schedule_kwargs)
            logger.info(f"Scheduled spider job {job.key}")
            return job.key
        except DuplicateJobError as e:
            logger.error(str(e))
        except:
            raise

    @dash_retry_decorator
    def is_running(self, jobkey, project_id=None):
        """
        Checks whether a job is running (or pending)
        """
        project = self.get_project(project_id)
        job = project.jobs.get(jobkey)
        if job.metadata.get('state') in ('running', 'pending'):
            return True
        return False

    @dash_retry_decorator
    def is_finished(self, jobkey, project_id=None):
        """
        Checks whether a job is running. if so, return close_reason. Otherwise return None.
        """
        project = self.get_project(project_id)
        job = project.jobs.get(jobkey)
        if job.metadata.get('state') == 'finished':
            return job.metadata.get('close_reason')
        return

    def wait_for(self, jobs_keys, interval=60, timeout=float('inf'), heartbeat=None):
        """Waits until all given jobs are not running anymore or until the
        timeout is reached, if a heartbeat is given it'll log an entry every
        heartbeat seconds (considering the interval), otherwise it'll log an
        entry every interval seconds.
        """
        if isinstance(jobs_keys, str):
            jobs_keys = [jobs_keys]
        if heartbeat is None or heartbeat < interval:
            heartbeat = interval
        still_running = dict((key, True) for key in jobs_keys)
        time_waited, next_heartbeat = 0, heartbeat
        while any(still_running.values()) and time_waited < timeout:
            time.sleep(interval)
            time_waited += interval
            for key in (k for k, v in still_running.items() if v):
                if self.is_running(key):
                    if time_waited >= next_heartbeat:
                        next_heartbeat += heartbeat
                        logger.info('{} still running'.format(key))
                    break
                else:
                    still_running[key] = False

    def on_start(self):
        self.workflow_loop_enabled = True

    def workflow_loop(self):
        pass

    def on_close(self):
        pass

    def __close(self):
        self.on_close()

    def run(self):
        self.on_start()
        self._run_loops()

    def _run_loops(self):
        while self.workflow_loop_enabled:
            try:
                if self.workflow_loop() and self.args.loop_mode:
                    time.sleep(self.args.loop_mode)
                else:
                    self.__close()
                    logger.info("No more tasks")
                    return
            except KeyboardInterrupt:
                logger.info("Bye")
                return



"""
Utility for cloning ScrapyCloud jobs
Features tagging of cloned from/to jobs (both source and destination) and avoids to clone source jobs already cloned.
By default cloned jobs are scheduled in the same project as source job. If --project-id is given, target project
is overriden.
"""
import logging
from shub_workflow.script import BaseScript

_LOG = logging.getLogger(__name__)
_LOG.setLevel(logging.INFO)


def _transform_cmd(job_cmd):
    if isinstance(job_cmd, list):
        return ' '.join(["'%s'" % cmd for cmd in job_cmd[1:]])

    return job_cmd


_COPIED_FROM_META = {'job_cmd':     ('cmd_args', _transform_cmd),
                     'units':       (None,       None),
                     'spider_args': ('job_args', None),
                     'tags':        ('add_tag',  None)}


def is_cloned(job):
    for tag in job.metadata.get('tags'):
        if tag.startswith('ClonedTo='):
            _LOG.warning(f'Job {job.key} already cloned. Skipped.')
            return True
    return False


def is_cloned_by_jobkey(jobkey, client):
    job = client.get_job(jobkey)
    return is_cloned(job)


def clone_job(job_key, client, units=None, default_project_id=None, extra_tags=None, job_params_hook=None):
    extra_tags = extra_tags or []
    job = client.get_job(job_key)

    spider = job.metadata.get('spider')

    job_params = dict()
    for key, (target_key, _) in _COPIED_FROM_META.items():

        if target_key is None:
            target_key = key

        job_params[target_key] = job.metadata.get(key)
        job_params.setdefault('add_tag', []).append(f'ClonedFrom={job_key}')
        job_params['add_tag'].extend(extra_tags)
        if units is not None:
            job_params['units'] = units

    if job_params_hook is not None:
        job_params_hook(job_params)

    for key, (target_key, transform) in _COPIED_FROM_META.items():

        target_key = target_key or key

        if transform is None:
            transform = lambda x: x

        job_params[target_key] = transform(job_params[target_key])

    project_id, spider_id, job_id = job_key.split('/')
    project = client.get_project(default_project_id or project_id)
    new_job = project.jobs.run(spider, **job_params)
    _LOG.info("Cloned %s to %s", job_key, new_job.key)
    jobtags = job.metadata.get('tags')
    jobtags.append(f'ClonedTo={new_job.key}')
    job.metadata.update({'tags': jobtags})


class CloneJobScript(BaseScript):

    flow_id_required = False

    @property
    def description(self):
        return __doc__

    def _clone_job(self, job_key):
        return clone_job(job_key, self.client, self.args.units, self.project_id, self.args.tag)

    def parse_project_id(self, args):
        project_id = super().parse_project_id(args)
        if project_id:
            return project_id
        if args.key:
            return args.key[0].split('/')[0]
        if args.tag_spider:
            return args.tag_spider.split('/')[0]

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument('--key', type=str, action='append', default=[],
                                    help='Target job key. Can be given multiple times. All must be in same project.')
        self.argparser.add_argument('--tag-spider', help='In format <project_id>/<tag>/<spider name>,'
                                    'clone given spider from given project id, by tag')
        self.argparser.add_argument('--units', help='Set number of units. Default is the same as cloned job.', type=int)
        self.argparser.add_argument('--tag', help='Additional tag added to the newly created jobs. Can be multiple.',
                                    default=[])

    def run(self):
        if self.args.key:
            keys = filter(lambda x: not is_cloned_by_jobkey(x, self.client), self.args.key)
        elif self.args.tag_spider:
            keys = []
            project_id, tag, spider = self.args.tag_spider.split('/')
            for job in self.client.get_project(project_id).jobs.iter(spider=spider, state=['finished'], has_tag=tag):
                if not is_cloned_by_jobkey(job['key'], self.client):
                    keys.append(job['key'])

        for job_key in keys:
            try:
                self._clone_job(job_key)
            except Exception as e:
                _LOG.error('Could not restart job %s: %s', job_key, e)


if __name__ == "__main__":
    script = CloneJobScript()
    script.run()



"""
Base script class for spiders crawl managers.
"""
import json

from shub_workflow.base import WorkFlowManager


class CrawlManager(WorkFlowManager):

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument('spider', help='Spider name')
        self.argparser.add_argument('--spider-args', help='Spider arguments dict in json format', default='{}')
        self.argparser.add_argument('--job-settings', help='Job settings dict in json format', default='{}')
        self.argparser.add_argument('--units', help='Set number of ScrapyCloud units for each job', type=int)

    def get_spider_args(self, override=None):
        spider_args = json.loads(self.args.spider_args)
        if override:
            spider_args.update(override)
        return spider_args

    def get_job_settings(self, override=None):
        job_settings = json.loads(self.args.job_settings)
        if override:
            job_settings.update(override)
        return job_settings

    def schedule_spider(self, spider_args_override=None, job_settings_override=None):
        spider_args = self.get_spider_args(spider_args_override)
        job_settings = self.get_job_settings(job_settings_override)
        spider_args['job_settings'] = job_settings
        return super().schedule_spider(self.args.spider, units=self.args.units, **spider_args)

    def workflow_loop(self):
        self.schedule_spider()
        return True
